/* Programmer: Steven Farago
 * Class:      EEL 5666
 * Description: This file is the primary location for the behaviors
                that control the Capture the Flag robots
 */


#include <hc11.h>
#include <mil.h>
#include <serial.h>
#include <analog.h>
#include <motor.h>

#include "ports.h"
#include "defs.h"
#include "time.c"
#include "utils.c"
#include "music.c"
#include "comm.c"

/* Motor routines place values in the motor locations for appropriate index
   of the behavior array
*/

/* Note:  Motor bugs in the ICC11 libs required that some of this code
   be written a bit backwards.  Any seemingly illogical motor commands in
   these direction definitions or in the behavior submodules are necessay
   in order to handle the bug.
*/

void go_forward(int behav, int speed)
{behavior[behav][MOTOR0]=speed;
 behavior[behav][MOTOR1]=speed;
}

void go_back(int behav, int speed)
{behavior[behav][MOTOR0]=-speed;
 behavior[behav][MOTOR1]=-speed;
}

void go_left(int behav, int speed)
{behavior[behav][MOTOR0]=speed;
 behavior[behav][MOTOR1]=-speed;
}

void go_right(int behav, int speed)
{behavior[behav][MOTOR0]=-speed;
 behavior[behav][MOTOR1]=speed;
}

void stop(int behav)
{behavior[behav][MOTOR0]=0;
 behavior[behav][MOTOR1]=0;
}

/* Used to calibrate the CdS cells.  Assumes the robot starts in a
 * location with a constant, known light level for all normal CdS cells
 */

void calibrate_cds()
{ncds_mid_nom=exp_analog(NCDS_MID);
 ncds_rt_nom=exp_analog(NCDS_RIGHT);
 ncds_lt_nom=exp_analog(NCDS_LEFT);
 pcds_up_nom=exp_analog(PCDS_UP);
 pcds_down_nom=exp_analog(PCDS_DOWN);
}

/* Initializes all software modules, global variables and behavior arrays
*/

void init_all()
{int i, j;

 init_motors();
 init_serial();
 init_analog();
 init_timers();
 init_music();
 init_rev_system();
 init_send_system();

 for(i=0; i<BEHAVIOR_COUNT; ++i)
   {behavior[i][DEFERRED_OP]=0;
    behavior[i][MODULUS]=1000;
    behavior[i][LAST_OP]=0;
    behavior[i][PRIORITY]=0;
   }

 done=0; had_flag=0;

 last_behavior=last_roam_behavior=ROAM_ACTIVE;
 set_timer_long(CONTROLLER_TIMER, 0x2DC, 0x6C00);

 turns=0; last_turn=LEFT;
 set_timer_long(BRADENBURG, 0x0098, 0x9680);

 cur_m0=cur_m1=0;
 target_m0=target_m1=0;
}


/* Handle a bump detection.  The robot will either back up or go forward
   (depending on the direction bumped) for a specified length of time, and
   then turn either right or left (again depending on the direction bumped)
*/

#define FL 3
#define FR 2
#define BK 1
void check_bump()
{int ops, last_op;
 if (ops=behavior[BUMP][DEFERRED_OP])
    {
     if (! check_timer(BUMP)) {return;}
     else
       {behavior[BUMP][DEFERRED_OP]--;
        last_op=behavior[BUMP][LAST_OP];
        switch (ops)
         {case 3 : if (last_op==FL || last_op==FR)
                    {go_back(BUMP, MED);}
                   else {go_forward(BUMP, MED);}
                   set_timer_long(BUMP,0x0010,0xC000);
                   break;

          case 2 : if (last_op==FL) {go_right(BUMP, MED);}
                   if (last_op==FR) {go_left(BUMP, MED);}
                   if (last_op==BK) {go_left(BUMP, MED);}
                   set_timer_long(BUMP,0x0010,0xC000);
                   break;
         }
       }
    }
 else
    {if ((bump_BK<128) || (bump_FL<128) || (bump_FR<128))
        {if (bump_BK<128) {behavior[BUMP][LAST_OP]=BK;}
         if (bump_FL<128) {behavior[BUMP][LAST_OP]=FL;}
         if (bump_FR<128) {behavior[BUMP][LAST_OP]=FR;}
         behavior[BUMP][DEFERRED_OP]=3;
         behavior[BUMP][PRIORITY]=priority[BUMP];
        }
     else
        {behavior[BUMP][DEFERRED_OP]=0; behavior[BUMP][PRIORITY]=0;}
    }
} 

/* Handles collision avoidance.  Triggered if any Sharps read above the
   threshold value.
*/

void IR_avoidance()
{int threshold=124;

 if ((ir_FL>threshold) || (ir_FR>threshold) ||
     (ir_LL>threshold) || (ir_RR>threshold))

   {
/*    if ((ir_LL>threshold) && (ir_RR>threshold)) {go_forward(IR_AVOID, MED);}
*/
    if (ir_FL>threshold) {go_right(IR_AVOID, FAST);}
    if (ir_FR>threshold) {go_left(IR_AVOID, FAST);}
    if (ir_LL>threshold+4) {go_right(IR_AVOID, MED);}
    if (ir_RR>threshold+4) {go_left(IR_AVOID, MED);}
    if ((ir_LL>threshold) && (ir_RR>threshold)) 
       {go_forward(IR_AVOID, MED);}


    behavior[IR_AVOID][PRIORITY]=priority[IR_AVOID];
   }
 else
   {behavior[IR_AVOID][PRIORITY]=0;}
}
 
/* Attempts to avoid the Braitenburg (pardon the spelling in the code) trap.
   The robot is continually counting the number of times it changes
   direction.  If the count exceeds the max time variable in a given time
   interval, this function instructs the robot to turn about 180 degrees
*/

void bradenburg()
{int ops;
 int max_turns=3;
 if (check_timer(BRADENBURG))
    {
     if (behavior[BRADENBURG][DEFERRED_OP])
        {set_timer_long(BRADENBURG, 0x0098, 0x9680);
         turns=0;
         behavior[BRADENBURG][DEFERRED_OP]=0;
         behavior[BRADENBURG][PRIORITY]=0;
        }
     else
       {if (turns>max_turns)
           {
            behavior[BRADENBURG][DEFERRED_OP]=1;
            behavior[BRADENBURG][PRIORITY]=priority[BRADENBURG];
            go_left(BRADENBURG, MED);
            set_timer_long(BRADENBURG, 0x0020, 0x0000);
           }
        else
           {turns=0;
            set_timer_long(BRADENBURG, 0x0098, 0x9680);
           }
       }
    }
 else {return;}
}

/* Instructs the robot to back up and turn.  Seems to have some bugs.
 * Similar operation to the check_bump routine
 */

void turn_180()
{int ops;

  if (ops=behavior[TURN_180][DEFERRED_OP])
    {if (! check_timer(TURN_180)) {return;}
     else
       {behavior[TURN_180][DEFERRED_OP]--;
         switch (ops)
         {case 3 : go_back(TURN_180, FAST);
                   set_timer_long(TURN_180,0x0010,0xC000);
                   break;

          case 2 : go_right(TURN_180, MED);
                   set_timer_long(TURN_180,0x0017,0x0000);
                   break;
         }
      }
    }
  else
    {if (turn180_active)
      {turn180_active=0;
       behavior[TURN_180][DEFERRED_OP]=3;
       behavior[TURN_180][PRIORITY]=priority[TURN_180];
      }
     else
      {behavior[TURN_180][DEFERRED_OP]=0;
       behavior[TURN_180][PRIORITY]=0;
      }
    }
 }

/* Disabled the robot is a disable code is detected from the digital
 * IR comm system
 */
void be_disabled()
{
 if (behavior[DISABLED][DEFERRED_OP])
    {if ((cur_m0==0) && (cur_m1==0))
        {switch(random(3))
          {case 0 : play_song(jeopardy); break;
           case 1 : play_song(lullabye); break;
           case 2 : play_song(dirge); break;
          }
         behavior[DISABLED][DEFERRED_OP]=0;
         behavior[DISABLED][PRIORITY]=0;
        }
     else {return;}
   }
 else {
       if (check_disable())
          {behavior[DISABLED][PRIORITY]=200;
           behavior[DISABLED][DEFERRED_OP]=1;
           stop(DISABLED);
          }
      }
}

/* Fires the disable gun.  Sends a disable code with the IR comm system
   and make the sound of a gun shooting */

void shoot_disable_gun()
{
  {if (! see_flag)
   if (check_timer(SHOOT))
    {IR_LEDS=0x00;
     send_message(moyer_code);
     gun_sound(1);
     send_off();
     IR_LEDS=FRONT_FOUR;
     set_timer_long(SHOOT, 0x01C9, 0xC380);
    }
  }
}

/* Implements IR following.  Function triggered to activity if any of the
   sharps register above a threshold value.  Robot is then instructed to
   turn (in the appropriate direction) until function becomes inactive
*/
void IR_follow()
{int threshold=110;
 if ((ir_FL>threshold) || (ir_FR>threshold) || (ir_LL>threshold) || (ir_RR>threshold))

   {behavior[IR_FOLLOW][PRIORITY]=priority[IR_FOLLOW];

    if ((ir_FR>(threshold+9)) || (ir_FL>(threshold+9)))
       {go_forward(IR_FOLLOW,MED); return;}

    if ((ir_RR>threshold) || (ir_LL>threshold))
       {if (ir_RR>ir_LL) {go_right(IR_FOLLOW,MED); return;}
        else {go_left(IR_FOLLOW,MED); return;}
       }

    go_forward(IR_FOLLOW,MED);
   }
 else {behavior[IR_FOLLOW][PRIORITY]=0;}
}

/* Similar to IR following routine.  Uses CDS cells to seek out light
   sources.  The function is theoretically only triggered to operation by
   polarized light
*/

void seek_flag()
{int threshold=55;
 if (see_flag)
  {if ((ncds_LT>threshold) || (ncds_RT>threshold) || (ncds_MID>threshold))
       {behavior[SEEK_FLAG][PRIORITY]=priority[SEEK_FLAG];

      if (ncds_MID>threshold) {go_forward(SEEK_FLAG,SLOW); return;}

      if ((ncds_LT>threshold) || (ncds_RT>threshold))
         {if (ncds_RT>threshold) {go_right(SEEK_FLAG, SLOW); return;}
          else {go_left(SEEK_FLAG, SLOW); return;}
         }

      go_forward(SEEK_FLAG,MED);

/*
        switch (max(ncds_MID, ncds_LT, ncds_RT))
          {case 1 : if (ncds_MID>140) {go_forward(SEEK_FLAG, SLOW);}
                    else {go_forward(SEEK_FLAG, MED); return;}
           case 2 : go_left(SEEK_FLAG, SLOW); return;
           case 3 : go_right(SEEK_FLAG, SLOW); return;
           default : go_forward(SEEK_FLAG, MED);
          }
        return;
*/

       }


  }
    else {behavior[SEEK_FLAG][PRIORITY]=0;}
   
 return;
}

/* If the robot has stolen the flag and is near its home base, then this
 * routine ends the game (by setting the done variable)
 */

void game_over()
{if ((had_flag==TRUE) && (see_flag==FRIENDLY) &&
     ((bump_FL<128) || (bump_FR<128) || (ir_FL>120) | (ir_FR>120)))
    {stop(GAME_OVER);
     behavior[GAME_OVER][PRIORITY]=199;
     if ((cur_m0==0) && (cur_m1==0)) {done=1;}
    }
}

/* Sets up the default behavior (go straight) */
void defa()
{behavior[DEFAULT][PRIORITY]=priority[DEFAULT];
 behavior[DEFAULT][MOTOR0]=MED;
 behavior[DEFAULT][MOTOR1]=MED;

}


/* Utility function for setting the target value of the motors.  Also
 * keeps track of the number of direction changes
 */

void set_motors(int m0, int m1)
{int cur_turn;
 if ((m0+m1)==0)
    {if (m1>m0) {cur_turn=LEFT;} else {cur_turn=RIGHT;}
     if ((cur_turn+last_turn)==0) {++turns;}
     last_turn=cur_turn;
    }
target_m0=m0;
target_m1=m1;
}

/* Handles motor smoothing */
void update_motors()
{int last0, last1;
 last0=cur_m0; last1=cur_m1;
 cur_m0=((cur_m0*15)+target_m0)/16;

 cur_m1=((cur_m1*15)+target_m1)/16;

 if ((cur_m0==last0) && (cur_m0<target_m0)) {++cur_m0;}
 if ((cur_m0==last0) && (cur_m0>target_m0)) {--cur_m0;}

 if ((cur_m1==last1) && (cur_m1<target_m1)) {++cur_m1;}
 if ((cur_m1==last1) && (cur_m1>target_m1)) {--cur_m1;}


 motor(0, cur_m0);
 motor(1, cur_m1);

/*
printf("m0 %d   m1 %d \n",cur_m0,cur_m1);
*/ 
}

/* This function defines the high level behaviors by specifing interactions
   among the submodules.
*/

void set_behavior(int dominant_behavior)
{
 
 switch (dominant_behavior)
     {case (ROAM_ACTIVE) : IR_LEDS=FRONT_FOUR;
                           priority[DEFAULT]=20;
                           priority[IR_AVOID]=175;
                           priority[BRADENBURG]=185;
                           priority[BUMP]=180;
                           priority[IR_FOLLOW]=0;
                           priority[SEEK_FLAG]=150;
                           break;
      
      case (ROAM_PASSIVE): IR_LEDS=0x00;
                           priority[DEFAULT]=20;
                           priority[IR_AVOID]=0;
                           priority[BRADENBURG]=180;
                           priority[BUMP]=175;
                           priority[IR_FOLLOW]=185;
                           priority[SEEK_FLAG]=150;
                           break;
     
      case (SEEK_MY_FLAG): IR_LEDS=0x00;
                           priority[DEFAULT]=20;
                           priority[IR_AVOID]=0;
                           priority[BRADENBURG]=175;
                           priority[BUMP]=185;
                           priority[IR_FOLLOW]=0;
                           priority[SEEK_FLAG]=180;
                           break;

      case (SEEK_MY_HOME): IR_LEDS=0x00;
                           priority[DEFAULT]=20;
                           priority[IR_AVOID]=0;
                           priority[BRADENBURG]=175;
                           priority[BUMP]=185;
                           priority[IR_FOLLOW]=0;
                           priority[SEEK_FLAG]=180;
                           break;

     case (WRONG_WAY)   :  IR_LEDS=0x00;
                           if (last_behavior!=WRONG_WAY) {turn180_active=1;}
                           priority[DEFAULT]=20;
                           priority[IR_AVOID]=180;
                           priority[BRADENBURG]=182;
                           priority[BUMP]=190;
                           priority[IR_FOLLOW]=0;
                           priority[SEEK_FLAG]=0;
                           priority[TURN_180]=0;
                           break;
   }
last_behavior=dominant_behavior;
}


/* Use 200 as max priority */

/* This function uses sensor variables to set high level behaviors */
void controller()
{

 
 if ((see_flag==ENEMY) && (! got_flag)) {set_behavior(SEEK_MY_FLAG);
                                         return;}
 if ((see_flag==ENEMY) && (got_flag))   {set_behavior(WRONG_WAY);
                                         return;}
 if ((see_flag==FRIENDLY) && (! got_flag)) {set_behavior(WRONG_WAY);
                                            return;}
 if ((see_flag==FRIENDLY) && (got_flag))   {set_behavior(SEEK_MY_HOME);
                                            return;}
 
                                                          
 if (check_timer(CONTROLLER_TIMER))
     {
      last_roam_behavior=(last_roam_behavior+1) % 2;
      if (last_roam_behavior==ROAM_ACTIVE) {play_song(charge);}
      else {play_song(spy);}
      set_timer_long(CONTROLLER_TIMER, 0x0262, 0x5A00);
     }
 
set_behavior(last_roam_behavior);
  
return;
    
}


   



void main()
{int i, index, max;
 init_all();
 defa();
 IR_LEDS=FRONT_FOUR;
 play_song(revelie_short);
 calibrate_cds();
 while(! done)
   {
    update_sensors();
    controller();
    check_bump();
    turn_180(); 

  
    IR_avoidance();

    IR_follow(); 
    be_disabled();

   
    bradenburg();
  

    seek_flag();


   
    shoot_disable_gun();
    game_over(); 

    update_motors();
    
    index=0; max=0;
    for(i=0; i<BEHAVIOR_COUNT; ++i)
       {if (behavior[i][PRIORITY]>max) {max=behavior[i][PRIORITY];
                                        index=i;}
       }
        set_motors(behavior[index][MOTOR0], behavior[index][MOTOR1]);
  
  }
play_song(fight_song);
}


